﻿/*
   Copyright 2013 uraku.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;
using System.Xml.Serialization;

namespace AutomaticScreenshotoConverter
{
    public partial class MainForm : Form
    {
        //ファイルの書き込みを監視するウォッチャー
        private FileSystemWatcher FsWatcher = null;
        //監視中フラグ、監視中であればtrue
        private bool IsMonitoring = false;
        //Jpeg用のCodecInfo
        private ImageCodecInfo JpegInfo;
        //前回処理したファイル
        private String BeforeFilePath = "";
        //削除対象ファイル
        private List<String> DeleteList = new List<String>();
        //設定ファイル
        private readonly String ConfigFilePath = Path.Combine(System.Environment.CurrentDirectory, "config.xml");

        public MainForm()
        {
            InitializeComponent();
            JpegInfo = GetCodecInfo("image/jpeg");
            TextBoxLog.Text = "Started";
        }

        private void RadioConvertJPEG_CheckedChanged(object sender, EventArgs e)
        {
            //JPEG用のコントロールを有効にする
            LabelJpegQuality.Enabled = true;
            NumericJpegQuality.Enabled = true;
        }

        private void RadioConvertPNG_CheckedChanged(object sender, EventArgs e)
        {
            //JPEG用のコントロールを無効にする
            LabelJpegQuality.Enabled = false;
            NumericJpegQuality.Enabled = false;
        }

        private void ButtonSourceFolderBrows_Click(object sender, EventArgs e)
        {
            //ソースフォルダー選択
            FolderBrowserDialog fbDialog = new FolderBrowserDialog();
            fbDialog.RootFolder = Environment.SpecialFolder.Desktop;
            fbDialog.ShowNewFolderButton = true;
            if (fbDialog.ShowDialog(this) == DialogResult.OK)
            {
                TextBoxSourceFolder.Text = fbDialog.SelectedPath;
            }
        }

        private void ButtonDestFolderBrows_Click(object sender, EventArgs e)
        {
            //コピー先フォルダー選択
            FolderBrowserDialog fbDialog = new FolderBrowserDialog();
            fbDialog.RootFolder = Environment.SpecialFolder.Desktop;
            fbDialog.ShowNewFolderButton = true;
            if (fbDialog.ShowDialog(this) == DialogResult.OK)
            {
                TextBoxDestFolder.Text = fbDialog.SelectedPath;
            }
        }

        /// <summary>
        /// フォルダー監視開始・終了処理
        /// </summary>
        private void ButtonMonitoring_Click(object sender, EventArgs e)
        {
            if (this.IsMonitoring)
            {
                //フォルダー監視終了処理
                WriteLog(Message.LogExitNow);
                FsWatcher.EnableRaisingEvents = false;
                FsWatcher.Dispose();
                FsWatcher = null;
                MonitoringUiModeEnd();
                this.IsMonitoring = false;
                if (CheckDeleteSource.Checked)
                {
                    DeleteSourceFiles();
                }
                WriteLog(Message.LogExit);
            }
            else
            {
                //未設定でないか
                if (!(IsConfigured()))
                    return;
                //設定したフォルダが存在しソースとコピー先が同じでないか
                if (!(IsValidFolder()))
                    return;
                //フォルダー監視の準備
                FsWatcher = new FileSystemWatcher(TextBoxSourceFolder.Text);
                FsWatcher.Filter = ""; //とりあえず全ファイル見る
                FsWatcher.IncludeSubdirectories = false; //サブディレクトリは見ない
                FsWatcher.SynchronizingObject = this;
                FsWatcher.Created += new System.IO.FileSystemEventHandler(FsWatcher_Changed);
                //フォルダー監視を開始する
                FsWatcher.EnableRaisingEvents = true;
                MonitoringUiMode();
                this.IsMonitoring = true;
                WriteLog(Message.LogStart);
            }
        }

        /// <summary>
        /// 設定が行われているかを確認。
        /// /// このメソッドは内部でMessageBoxを表示する。
        /// </summary>
        /// <returns>問題が無ければtrueを返す</returns>
        private bool IsConfigured()
        {
            if ((TextBoxSourceFolder.TextLength < 1) || (TextBoxDestFolder.TextLength < 1))
            {
                MessageBox.Show(Message.NotConfig, "", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return false;
            }
            return true;
        }

        /// <summary>
        /// 指定したフォルダーが存在しているか。競合していないかを確認。
        /// このメソッドは内部でMessageBoxを表示する。
        /// </summary>
        /// <returns>問題が無ければtrueを返す</returns>
        private bool IsValidFolder()
        {
            if (!(Directory.Exists(TextBoxSourceFolder.Text)))
            {
                MessageBox.Show(Message.SourceNotFound, Message.Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
            if (!(Directory.Exists(TextBoxDestFolder.Text)))
            {
                MessageBox.Show(Message.DestNotFound, Message.Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
            if (TextBoxSourceFolder.Text == TextBoxDestFolder.Text)
            {
                MessageBox.Show(Message.SdSameLocation, Message.Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
            return true;
        }


        /// <summary>
        /// 監視フォルダーでのファイル作成イベントハンドラー
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        private void FsWatcher_Changed(Object source, FileSystemEventArgs e)
        {
            if (e.ChangeType == WatcherChangeTypes.Created)
            {
                //前回と同じファイルパスであれば無視
                if (BeforeFilePath == e.FullPath)
                    return;
                ImageConvertExec(e.FullPath);
                BeforeFilePath = e.FullPath;
            }
        }

        /// <summary>
        /// 画像変換処理。設定済みの項目にしたがって画像変換や元ファイルの削除を行う。
        /// </summary>
        /// <param name="sourcePath">ソースになるファイルのフルパス</param>
        private void ImageConvertExec(String sourcePath)
        {
            //元ファイルのファイル形式確認
            String ext = Path.GetExtension(sourcePath);
            if (RadioSourceBMP.Checked)
            {
                if ((ext != ".bmp") && (ext != ".BMP"))
                    return;
            }
            else if (RadioSourcePNG.Checked)
            {
                if ((ext != ".png") && (ext != ".PNG"))
                    return;
            }
            else if (RadioSourceJPEG.Checked)
            {
                if ((ext != ".jpg") && (ext != ".JPG") && (ext != ".jpeg") && (ext != ".JPEG"))
                    return;
            }
            WriteLog("input:" + sourcePath);
            //元ファイルの書き込み完了を待機
            if (!(WaitForWriteComplete(sourcePath)))
            {
                WriteLog(Message.LogInputReadError + sourcePath);
                return;
            }
            // 新しい名前を付ける
            String newName = null;
            if (RadioNameDateTime.Checked)
                newName = File.GetLastWriteTime(sourcePath).ToString("s").Replace(':', '-');
            else
                newName = Path.GetFileNameWithoutExtension(sourcePath);

            String newExt = null;
            if (RadioConvertPNG.Checked)
                newExt = ".png";
            else
                newExt = ".jpg";

            int i = 0;
            String CheckName = newName;
            while (!(CheckDestinationConflictName(CheckName + newExt)))
            {
                i++;
                CheckName = newName + "_" + i;
                if (i > 998)
                {
                    WriteLog(Message.LogConflictDest);
                    return;
                }
            }
            newName = CheckName + newExt;
            String newFileFullPath = Path.Combine(TextBoxDestFolder.Text, newName);

            //画像の変換と保存
            using (Bitmap sourceBmp = new Bitmap(sourcePath))
            {
                if (RadioConvertPNG.Checked)
                {
                    sourceBmp.Save(newFileFullPath, ImageFormat.Png);
                    WriteLog(Message.LogOutput + newName);
                }
                else
                {
                    EncoderParameters eps = new EncoderParameters(1);
                    int quality = (int)NumericJpegQuality.Value;
                    EncoderParameter ep = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, quality);
                    eps.Param[0] = ep;
                    ImageCodecInfo ici = JpegInfo;
                    if (ici == null)
                    {
                        WriteLog(Message.LogInternalError);
                        sourceBmp.Dispose();
                        return;
                    }
                    sourceBmp.Save(newFileFullPath, ici, eps);
                    WriteLog(Message.LogOutput + newFileFullPath);
                }
            }
            if (CheckDeleteSource.Checked)
            {
                DeleteList.Add(sourcePath);
            }

        }

        /// <summary>
        /// 変換した元ファイルをまとめて削除する
        /// </summary>
        private void DeleteSourceFiles()
        {
            foreach (String filePath in DeleteList)
            {
                if (File.Exists(filePath))
                {
                    File.Delete(filePath);
                    WriteLog(Message.LogDeleted + filePath);
                }
            }
        }

        /// <summary>
        /// 新しく付けようとしている名前が
        /// 保存先に既に存在していないか確認する。
        /// </summary>
        /// <param name="newFileName">新しいファイル名</param>
        /// <returns>存在していなければtrue</returns>
        private bool CheckDestinationConflictName(String newFileName)
        {
            String checkPath = Path.Combine(TextBoxDestFolder.Text, newFileName);
            if (File.Exists(checkPath))
                return false;
            return true;
        }

        /// <summary>
        /// MimeTypeからImageCodecInfoを取得
        /// </summary>
        /// <param name="mineType">MIME TYPE</param>
        /// <returns>ImageEncorderInfo</returns>
        private ImageCodecInfo GetCodecInfo(string mineType)
        {
            ImageCodecInfo[] encorders = ImageCodecInfo.GetImageEncoders();
            foreach (ImageCodecInfo info in encorders)
                if (info.MimeType == mineType)
                    return info;
            return null;
        }

        /// <summary>
        /// 対象のファイルが書き込み中でないか確認する。
        /// 書き込み中であると判断した場合最大100回まで待機する。
        /// </summary>
        /// <param name="fullpath">検査するファイル</param>
        private bool WaitForWriteComplete(String fullpath)
        {
            System.Threading.Thread.Sleep(1200);
            for (int i = 0; i < 100; i++)
            {
                try
                {
                    if (File.Exists(fullpath))
                    {
                        File.Move(fullpath, fullpath);
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                catch
                {
                    System.Threading.Thread.Sleep(800);
                }
            }
            return false;
        }

        /// <summary>
        /// UIを監視状態にする。
        /// </summary>
        private void MonitoringUiMode()
        {
            LabelSorceForder.Enabled = false;
            LabelSourceFormat.Enabled = false;
            LabelDestFolder.Enabled = false;
            LabelFileName.Enabled = false;
            LabelConvertFormat.Enabled = false;
            LabelJpegQuality.Enabled = false;

            TextBoxSourceFolder.Enabled = false;
            RadioSourceBMP.Enabled = false;
            RadioSourcePNG.Enabled = false;
            RadioSourceJPEG.Enabled = false;
            TextBoxDestFolder.Enabled = false;
            RadioNameKeep.Enabled = false;
            RadioNameDateTime.Enabled = false;
            RadioConvertJPEG.Enabled = false;
            RadioConvertPNG.Enabled = false;
            NumericJpegQuality.Enabled = false;
            CheckDeleteSource.Enabled = false;

            ButtonSourceFolderBrows.Enabled = false;
            ButtonDestFolderBrows.Enabled = false;

            ButtonMonitoring.Text = Message.ButtonEnd;
        }

        /// <summary>
        /// UIの監視状態を終了させる
        /// </summary>
        private void MonitoringUiModeEnd()
        {
            LabelSorceForder.Enabled = true;
            LabelSourceFormat.Enabled = true;
            LabelDestFolder.Enabled = true;
            LabelFileName.Enabled = true;
            LabelConvertFormat.Enabled = true;

            TextBoxSourceFolder.Enabled = true;
            RadioSourceBMP.Enabled = true;
            RadioSourcePNG.Enabled = true;
            RadioSourceJPEG.Enabled = true;
            TextBoxDestFolder.Enabled = true;
            RadioNameKeep.Enabled = true;
            RadioNameDateTime.Enabled = true;
            RadioConvertJPEG.Enabled = true;
            RadioConvertPNG.Enabled = true;
            if (RadioConvertJPEG.Checked)
            {
                LabelJpegQuality.Enabled = true;
                NumericJpegQuality.Enabled = true;
            }
            else
            {
                LabelJpegQuality.Enabled = false;
                NumericJpegQuality.Enabled = false;
            }
            CheckDeleteSource.Enabled = true;

            ButtonSourceFolderBrows.Enabled = true;
            ButtonDestFolderBrows.Enabled = true;

            ButtonMonitoring.Text = Message.ButtonMonitoring;
        }

        /// <summary>
        /// ログウィンドウにメッセージを書き込む。
        /// </summary>
        /// <param name="message">メッセージ</param>
        private void WriteLog(String message)
        {
            Monitor.Enter(message);
            try
            {
                StringBuilder builder = new StringBuilder();
                builder.Append(TextBoxLog.Text);
                builder.Append("\r\n");
                builder.Append(message);
                TextBoxLog.Text = builder.ToString();
                //末尾にスクロール
                TextBoxLog.SelectionStart = TextBoxLog.Text.Length;
                TextBoxLog.Focus();
                TextBoxLog.ScrollToCaret();
            }
            finally
            {
                Monitor.Exit(message);
            }
        }


        private void ButtonClearLog_Click(object sender, EventArgs e)
        {
            //ログをクリア
            TextBoxLog.Text = "";
        }

        /// <summary>
        /// 設定値の保存
        /// </summary>
        private void SaveConfig()
        {
            Configs config = new Configs();
            config.sourceForder = TextBoxSourceFolder.Text;
            if (RadioSourceBMP.Checked)
                config.sourceFormat = SourceFromat.BMP;
            else if (RadioSourcePNG.Checked)
                config.sourceFormat = SourceFromat.PNG;
            else
                config.sourceFormat = SourceFromat.PNG;
            config.destForder = TextBoxDestFolder.Text;
            if (RadioNameKeep.Checked)
                config.nameChanged = NameChange.KEEP;
            else
                config.nameChanged = NameChange.DATETIME;
            if (RadioConvertPNG.Checked)
                config.convertFormat = ConvertFromat.PNG;
            else
                config.convertFormat = ConvertFromat.JPEG;
            config.jpegQuality = (int)NumericJpegQuality.Value;
            if (CheckDeleteSource.Checked)
                config.sourceDelete = true;
            else
                config.sourceDelete = false;
            XmlSerializer serializer = new XmlSerializer(typeof(Configs));
            using (FileStream fs = new FileStream(ConfigFilePath, FileMode.Create))
            {
                serializer.Serialize(fs, config);
            }
        }

        /// <summary>
        /// 設定値の読み込み
        /// </summary>
        private void RestoreConfig()
        {
            if (!(File.Exists(ConfigFilePath)))
                return;
            XmlSerializer serializer = new XmlSerializer(typeof(Configs));
            Configs config;
            using (FileStream fs = new FileStream(ConfigFilePath, FileMode.Open))
            {
                config = (Configs)serializer.Deserialize(fs);
            }
            TextBoxSourceFolder.Text = config.sourceForder;
            if (config.sourceFormat == SourceFromat.BMP)
            {
                RadioSourceBMP.Checked = true;
                RadioSourcePNG.Checked = false;
                RadioSourceJPEG.Checked = false;
            }
            else if (config.sourceFormat == SourceFromat.PNG)
            {
                RadioSourceBMP.Checked = false;
                RadioSourcePNG.Checked = true;
                RadioSourceJPEG.Checked = false;
            }
            else
            {
                RadioSourceBMP.Checked = false;
                RadioSourcePNG.Checked = false;
                RadioSourceJPEG.Checked = true;
            }
            TextBoxDestFolder.Text = config.destForder;
            if (config.nameChanged == NameChange.KEEP)
            {
                RadioNameKeep.Checked = true;
                RadioNameDateTime.Checked = false;
            }
            else
            {
                RadioNameKeep.Checked = false;
                RadioNameDateTime.Checked = true;
            }
            if (config.convertFormat == ConvertFromat.PNG)
            {
                RadioConvertPNG.Checked = true;
                RadioConvertJPEG.Checked = false;
            }
            else
            {
                RadioConvertPNG.Checked = false;
                RadioConvertJPEG.Checked = true;
            }
            int quality = config.jpegQuality;
            if ((quality >= 50) && (quality <= 100))
                NumericJpegQuality.Value = quality;
            else
                NumericJpegQuality.Value = 95;
            if (config.sourceDelete)
                CheckDeleteSource.Checked = true;
            else
                CheckDeleteSource.Checked = false;
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            RestoreConfig();
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (this.IsMonitoring)
            {
                //フォルダー監視終了処理
                WriteLog(Message.LogExitNow);
                FsWatcher.EnableRaisingEvents = false;
                FsWatcher.Dispose();
                FsWatcher = null;
                MonitoringUiModeEnd();
                this.IsMonitoring = false;
                if (CheckDeleteSource.Checked)
                {
                    DeleteSourceFiles();
                }
                WriteLog(Message.LogExit);
            }
            //設定値を保存
            SaveConfig();
            Application.Exit();
        }

        private void Button_Info_Click(object sender, EventArgs e)
        {
            MessageBox.Show(Message.InfoText, Message.Info);
        }
    }
}
